from .module import Module, interact

class Handy(Module):
    def __init__(self, id, alias, device):
        Module.__init__(self, 'Handy', id, alias, device)

        self.fingers = {'index': 0.0, 'middle': 0.0, 'ring': 0.0, 'pinky': 0.0, 'thumb': 0.0, }

    def power_ratio(self, s, name):
        s = min(max(s, 0.0), 100.0)
        self.fingers[name] = s
        self._push_value(name,s)

    @property
    def index(self):
        return self.fingers["index"]

    @index.setter
    def index(self, s):
        self.power_ratio(s,"index")

    @property
    def middle(self):
        return self.fingers["middle"]

    @middle.setter
    def middle(self, s):
        self.power_ratio(s,"middle")

    @property
    def ring(self):
        return self.fingers["ring"]

    @ring.setter
    def ring(self, s):
        self.power_ratio(s,"ring")

    @property
    def pinky(self):
        return self.fingers["pinky"]

    @pinky.setter
    def pinky(self, s):
        self.power_ratio(s,"pinky")

    @property
    def thumb(self):
        return self.fingers["thumb"]

    @thumb.setter
    def thumb(self, s):
        self.power_ratio(s,"thumb")

    def control(self):
        def change_pos(index, middle, ring, pinky, thumb):
            self.index = index
            self.middle = middle
            self.ring = ring
            self.pinky = pinky
            self.thumb = thumb

        return interact(change_pos,
                        index=(0.0, 100.0, 1.0),
                        middle=(0.0, 100.0, 1.0),
                        ring=(0.0, 100.0, 1.0),
                        pinky=(0.0, 100.0, 1.0),
                        thumb=(0.0, 100.0, 1.0))



from .module import Module, interact
from copy import copy
import time

class Stepper(Module):
    # target modes
    _MODE_COMPLIANT = 11
    _MODE_POWER = 10
    _MODE_ROT_SPEED = 8
    _MODE_ROT_POSITION = 7
    _MODE_TRANS_SPEED = 6
    _MODE_TRANS_POSITION = 5
    # report modes
    _ROTATION_POSITION = 4
    _ROTATION_SPEED = 3
    _TRANSLATION_POSITION = 2
    _TRANSLATION_SPEED = 1
    _CURRENT = 0

    def __init__(self, id, alias, device):
        Module.__init__(self, 'Stepper', id, alias, device)
        self._config = [False] * (Stepper._MODE_COMPLIANT + 1)
        # default configs, enable compliant, power_mode, and rotation position report
        self._config[Stepper._MODE_ROT_POSITION] = True
        self._config[Stepper._MODE_COMPLIANT] = True

        #configuration
        self._resolution = 200.0
        self._dimension = 0.0

        #targets
        self._target_rot_speed = 100.0
        self._target_rot_position = 0.0
        self._target_trans_speed = 0.0
        self._target_trans_position = 0.0

    def _convert_config(self):
        return int(''.join(['1' if c else '0' for c in self._config]), 2) # Table read reversly

    def bit(self, i, enable):
        self._config = self._config[:i] + () + self._config[i + 1:]

#************************** configurations *****************************

    def setToZero(self):
        self._push_value('reinit', None)

    @property
    def stepPerTurn(self):
        return self._resolution

    @stepPerTurn.setter
    def stepPerTurn(self, s):
        self._resolution = s
        self._push_value("resolution", s)

    @property
    def wheel_size(self):
        return self._dimension

    @wheel_size.setter
    def wheel_size(self, s):
        self._dimension = s
        self._push_value("dimension", s)

#************************** target modes *****************************

    # compliant
    @property
    def compliant(self):
        self._compliant

    @compliant.setter
    def compliant(self, enable):
        self._config[Stepper._MODE_COMPLIANT] = True if enable != 0  else False
        self._compliant = enable
        self._push_value('parameters', self._convert_config())
        time.sleep(0.01)

    # rotation speed
    @property
    def target_rot_speed(self):
        return self._target_rot_speed

    @target_rot_speed.setter
    def target_rot_speed(self, s):
        self._target_rot_speed = s
        self._push_value("target_rot_speed", s)

    # rotation position
    @property
    def target_rot_position(self):
        if (self._config[Stepper._MODE_ROT_POSITION] != True):
            print("rotation position mode is not enabled in the module please use 'device.module.rot_position_mode(True)' to enable it")
        return self._target_rot_position

    @target_rot_position.setter
    def target_rot_position(self, s):
        if (self._config[Stepper._MODE_ROT_POSITION] != True):
            print("rotation position mode is not enabled in the module please use 'device.module.rot_position_mode(True)' to enable it")
        self._target_rot_position = s
        self._push_value("target_rot_position", s)

    def rot_position_mode(self, enable):
        self._config[Stepper._MODE_ROT_POSITION] = True if enable != 0  else False
        if (enable == True) :
            self._config[Stepper._MODE_TRANS_POSITION] = False
        self._push_value('parameters', self._convert_config())
        time.sleep(0.01)

    # translation speed
    @property
    def target_trans_speed(self):
        return self._target_trans_speed

    @target_trans_speed.setter
    def target_trans_speed(self, s):
        if (self._dimension == 0) :
            print("you have to setup a wheel_size before using translation command")
        self._target_trans_speed = s
        self._push_value("target_trans_speed", s)

    # translation position
    @property
    def target_trans_position(self):
        if (self._config[Stepper._MODE_TRANS_POSITION] != True):
            print("translation speed mode is not enabled in the module please use 'device.module.trans_speed_mode(True)' to enable it")
        return self._target_trans_position

    @target_trans_position.setter
    def target_trans_position(self, s):
        if (self._config[Stepper._MODE_TRANS_POSITION] != True):
            print("translation speed mode is not enabled in the module please use 'device.module.trans_speed_mode(True)' to enable it")
        self._target_trans_position = s
        self._push_value("target_trans_position", s)

    def trans_position_mode(self, enable):
        if (self._dimension == 0) :
            print("you have to setup a wheel_size before using translation command")
        self._config[Stepper._MODE_TRANS_POSITION] = True if enable != 0  else False
        if (enable == True) :
            self._config[Stepper._MODE_ROT_POSITION] = False
        self._push_value('parameters', self._convert_config())
        time.sleep(0.01)

#************************** controls and updates *****************************

    def _update(self, new_state):
        Module._update(self, new_state)

    def control(self):
        def change_config(compliant_mode, rot_speed, rot_position_mode, rot_position, trans_speed, trans_position_mode, trans_position):
            # target mode
            self.compliant = compliant_mode
            self.target_rot_speed = rot_speed

            self.rot_position_mode(rot_position_mode)
            if (rot_position_mode) :
                self.target_rot_position = rot_position

            self.target_trans_speed = trans_speed

            self.trans_position_mode(trans_position_mode)
            if (trans_position_mode) :
                self.target_trans_position = trans_position


        w = interact(change_config,
                        compliant_mode = self._config[Stepper._MODE_COMPLIANT],
                        rot_speed = (-700, 700, 1),
                        rot_position_mode = self._config[Stepper._MODE_ROT_POSITION],
                        rot_position = (-360.0, 360.0, 1.0),
                        trans_speed = (-1000, 1000, 1),
                        trans_position_mode = self._config[Stepper._MODE_TRANS_POSITION],
                        trans_position = (-1000, 1000, 1))



from .gate import Gate
from .generic_motor_controller import GenericMotorController
from .distance import Distance
from .l0_gpio import GPIO
from .dc_motor import DCMotor
from .dxl import DynamixelMotor
from .state import State
from .angle import Angle
from .servo import Servo
from .color import Color
from .handy import Handy
from .imu import Imu
from .stepper import Stepper
from .power_switch import PowerSwitch
from .light_sensor import LightSensor
from .controlled_motor import ControlledMotor
from .void import Void
from .load import Load
from .voltage import Voltage
from .unknown import Unknown


__all__ = [
    'name2mod',
    'Gate',
    'GenericMotorController',
    'DynamixelMotor',
    'Distance',
    'State',
    'GPIO',
    'Angle',
    'Servo',
    'Color',
    'DCMotor',
    'Handy',
    'Imu' ,
    'Stepper',
    'PowerSwitch',
    'LightSensor',
    'ControlledMotor',
    'Void',
    'Load',
    'Voltage',
    'Unknown'
]

name2mod = {
    'Gate': Gate,
    'GenericMotor': GenericMotorController,
    'DynamixelMotor': DynamixelMotor,
    'DistanceSensor': Distance,
    'GPIO': GPIO,
    'DCMotor': DCMotor,
    'State': State,
    'Angle': Angle,
    'Servo': Servo,
    'Color': Color,
    'Handy': Handy,
    'Imu': Imu,
    'Stepper' : Stepper,
    'PowerSwitch' : PowerSwitch,
    'LightSensor' : LightSensor,
    'ControlledMotor' : ControlledMotor,
    'Void' : Void,
    'Load' : Load,
    'Voltage' : Voltage,
    'Unknown' : Unknown
}



from collections import defaultdict, namedtuple

import logging
import time

try:
    from ipywidgets import interact
    from ipywidgets import widgets
except ImportError:
    def interact(*args, **kwargs):
        msg = 'You first have to install ipywidgets to use the control widgets.'
        logging.getLogger(__name__).warning(msg)
        return None

    def widgets(*args, **kwargs):
        msg = 'You first have to install ipywidgets to use the control widgets.'
        logging.getLogger(__name__).warning(msg)
        return None

Event = namedtuple('Event', ('name', 'old_value', 'new_value'))


class Module(object):
    possible_events = set()

    def __init__(self,
                 type, id, alias,
                 device):
        self.id = id
        self.type = type
        self.alias = alias
        self.refresh_freq = 0.0
        self._delegate = device
        self._value = None
        self._cb = defaultdict(list)
        self._led = False
        self._L0_temperature = None
        self._L0_voltage = None
        self._firmware_revision = "Unknown"
        self._uuid = [0, 0, 0]
        self._killed = False
        self._last_update = time.time()

    def __repr__(self):
        return ('<{self.type} '
                'alias="{self.alias}" '
                'id={self.id}>'.format(self=self))

    def _update(self, new_state):
        if ((time.time() - self._last_update) != 0):
            self.refresh_freq = ((200.0 * self.refresh_freq) + (1.0 / (time.time() - self._last_update))) / 201.0
            self._last_update = time.time()
        if 'L0_temperature' in new_state:
            self._L0_temperature = new_state['L0_temperature']
        if 'L0_voltage' in new_state:
            self._L0_voltage = new_state['L0_voltage']
        if 'revision' in new_state:
            self._firmware_revision = new_state['revision']
        if 'uuid' in new_state:
            self._uuid = new_state['uuid']

    def _kill(self):
        self._killed = True
        print ("module", self.alias, "have been excluded from the network due to no responses.")

    def _push_value(self, key, new_val):
        if (self._killed) :
            print("module", self.alias,"is excluded.")
        else :
            if isinstance(new_val, float) :
                self._delegate.update_cmd(self.alias, key, float(str("%.3f" % new_val)))
            else :
                self._delegate.update_cmd(self.alias, key, new_val)

    def _push_data(self, key, new_val, data):
        if (self._killed) :
            print("module", self.alias,"is excluded.")
        else :
            self._delegate.update_data(self.alias, key, new_val, data)

    @property
    def L0_temperature(self):
        self._push_value('L0_temperature', "")
        time.sleep(0.03)
        return self._L0_temperature

    @property
    def L0_voltage(self):
        self._push_value('L0_voltage', "")
        time.sleep(0.03)
        return self._L0_voltage

    @property
    def firmware_revision(self):
        self._push_value('revision', "")
        time.sleep(0.03)
        return self._firmware_revision

    @property
    def uuid(self):
        self._push_value('uuid', "")
        time.sleep(0.03)
        return self._uuid

    @property
    def led(self):
        return self._led

    @led.setter
    def led(self, state):
        if state != self._led:
            self._push_value('led', state)
            self._led = state


    def led_toggle(self):
        if self._led is True:
            self._push_value('led', False)
            self._led = False
        else:
            self._push_value('led', True)
            self._led = True

    def rename(self, name):
        # check if the string start with a number before sending
        self._push_value('rename', name)
        self.alias = name


    # Events cb handling

    def add_callback(self, event, cb):
        if event not in self.possible_events:
            raise ValueError('Unknown callback: {} (poss={})'.format(event, self.possible_events))

        self._cb[event].append(cb)

    def remove_callback(self, event, cb):
        self._cb[event].remove(cb)

    def _pub_event(self, trigger, old_value, new_value):
        event = Event(trigger, old_value, new_value)

        for cb in self._cb[trigger]:
            cb(event)



from .module import Module


class Gate(Module):

    def __init__(self, id, alias, device):
        Module.__init__(self, 'Gate', id, alias, device)

    def _update(self, new_state):
        Module._update(self, new_state)

    @property
    def delay_ms(self):
        self._value

    @delay_ms.setter
    def delay_ms(self, s):
        self._value = s
        self._push_value("delay",s)

    def control(self):
        def delay(delay):
            self._value = delay

        return interact(delay, delay=(0, 100, 1))



from __future__ import division

from .module import Module, interact


class Servo(Module):
    def __init__(self, id, alias, device):
        Module.__init__(self, 'Servo', id, alias, device)
        self._max_angle = 180.0
        self._min_pulse = 0.0005
        self._max_pulse = 0.0015
        self._angle = 0.0

    @property
    def rot_position(self):
        return self._angle

    @rot_position.setter
    def rot_position(self, new_pos):
        self._angle = new_pos
        self._push_value('target_rot_position', new_pos)

    @property
    def max_angle(self):
        return self._max_angle

    @max_angle.setter
    def max_angle(self, new):
        self._max_angle = new
        param = [self._max_angle, self._min_pulse, self._max_pulse]
        self._push_value('parameters', param)

    @property
    def min_pulse(self):
        return self._min_pulse

    @min_pulse.setter
    def min_pulse(self, new):
        self._min_pulse = new
        param = [self._max_angle, self._min_pulse, self._max_pulse]
        self._push_value('parameters', param)

    @property
    def max_pulse(self):
        return self._max_pulse

    @max_pulse.setter
    def max_pulse(self, new):
        self._max_pulse = new
        param = [self._max_angle, self._min_pulse, self._max_pulse]
        self._push_value('parameters', param)

    def _update(self, new_state):
        Module._update(self, new_state)

    def control(self):
        def move(position):
            self.position = position

        return interact(move, position=(0, 180, 1))



from .module import Module


class Unknown(Module):

    def __init__(self, id, alias, device):
        Module.__init__(self, 'Unknown', id, alias, device)

    def _update(self, new_state):
        Module._update(self, new_state)



from .module import Module, interact
from .gpio import AnalogInputPin, DigitalInputPin, DigitalOutputPin, Pwm


class GPIO(Module):
    def __init__(self, id, alias, device):
        Module.__init__(self, 'GPIO', id, alias, device)

        self.analog_1 = AnalogInputPin('p1')

        self.digital_2 = DigitalOutputPin('p2', self)
        self.digital_3 = DigitalOutputPin('p3', self)
        self.digital_4 = DigitalOutputPin('p4', self)

        self.digital_5 = DigitalInputPin('p5')
        self.digital_6 = DigitalInputPin('p6')
        self.analog_7 = AnalogInputPin('p7')
        self.analog_8 = AnalogInputPin('p8')
        self.analog_9 = AnalogInputPin('p9')

    def _update(self, new_state):
        Module._update(self, new_state)

        if 'p1' in new_state:
            self.analog_1._update(new_state['p1'])
            self.digital_5._update(new_state['p5'])
            self.digital_6._update(new_state['p6'])
            self.analog_7._update(new_state['p7'])
            self.analog_8._update(new_state['p8'])
            self.analog_9._update(new_state['p9'])

    # notebook things
    def control(self):
        def change_pin(p2, p3, p4):
            self.digital_2._push(p2)
            self.digital_3._push(p3)
            self.digital_4._push(p4)

        return interact(change_pin,
                        p2=self.digital_2.is_high(),
                        p3=self.digital_3.is_high(),
                        p4=self.digital_4.is_high())
